#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <ctype.h>

#include "asm.h"
#include "evaluate.h"


// #define SUPPORT_IMPORT

#define BAD_REGISTER          0xffffffff

struct ASM {
  SYMBOL symbols[MAXSYMBOLS];
  U32 symbolcount;
  char errmsg[256];
  U32 totaloutputsize;
  unsigned char *output;
  char *currentline;
  U32 errorcount;
  ASMERROR errors[MAXERRORS];
};


static char *assemble_instruction(ASM *assm, EVALUATOR *eval, char *s, U32 adr, unsigned char *instruc, U32 *n);
static char *define_symbol(ASM *assm, EVALUATOR *eval, char *name, char *value, U32 flags);
static char *evaluate(ASM *assm, EVALUATOR *eval, char *s, S32 *value);
static char *constant_8bit(ASM *assm, EVALUATOR *eval, char *c, S32 *value, U32 *shift);
static char *shift(ASM *assm, EVALUATOR *eval, char *c, U32 *value, U32 *reg, U32 *shift);

static char *split_symbol_list(ASM *assm, char *p, char **list);

static char *get_until(char **src, char *out, char split);
static void tidy_instruction(char *p);
static U32 valid_symbol_char(char c);
static U32 valid_symbol_start_char(char c);
static char *split_in_4(char *s, char *arg1, char *arg2, char *arg3, char *arg4);
static U32 extract_register(char *r);
static U32 extract_condition(char *s, U32 *condition);
static void dummy_callback(char *e, LINE *l, U32 n);


#define _ADD        ('A' | ('D'<<8) | ('D'<<16))
#define _SUB        ('S' | ('U'<<8) | ('B'<<16))
#define _ADC        ('A' | ('D'<<8) | ('C'<<16))
#define _SBC        ('S' | ('B'<<8) | ('C'<<16))
#define _RSB        ('R' | ('S'<<8) | ('B'<<16))
#define _RSC        ('R' | ('S'<<8) | ('C'<<16))
#define _BIC        ('B' | ('I'<<8) | ('C'<<16))
#define _EOR        ('E' | ('O'<<8) | ('R'<<16))
#define _ORR        ('O' | ('R'<<8) | ('R'<<16))
#define _AND        ('A' | ('N'<<8) | ('D'<<16))

#define _MOV        ('M' | ('O'<<8) | ('V'<<16))
#define _MVN        ('M' | ('V'<<8) | ('N'<<16))

#define _CMP        ('C' | ('M'<<8) | ('P'<<16))
#define _CMN        ('C' | ('M'<<8) | ('N'<<16))
#define _TEQ        ('T' | ('E'<<8) | ('Q'<<16))
#define _TST        ('T' | ('S'<<8) | ('T'<<16))

#define _SWI        ('S' | ('W'<<8) | ('I'<<16))

#define _MUL        ('M' | ('U'<<8) | ('L'<<16))
#define _MLA        ('M' | ('L'<<8) | ('A'<<16))

#define _NOP        ('N' | ('O'<<8) | ('P'<<16))
#define _ADR        ('A' | ('D'<<8) | ('R'<<16))
#define _DCB        ('D' | ('C'<<8) | ('B'<<16))
#define _DCW        ('D' | ('C'<<8) | ('W'<<16))
#define _DCD        ('D' | ('C'<<8) | ('D'<<16))
#define _DCS        ('D' | ('C'<<8) | ('S'<<16))

#define _LDR        ('L' | ('D'<<8) | ('R'<<16))
#define _STR        ('S' | ('T'<<8) | ('R'<<16))
#define _LDM        ('L' | ('D'<<8) | ('M'<<16))
#define _STM        ('S' | ('T'<<8) | ('M'<<16))

#define _NV         ('N' | ('V'<<8))
#define _AL         ('A' | ('L'<<8))
#define _LE         ('L' | ('E'<<8))
#define _GT         ('G' | ('T'<<8))
#define _LT         ('L' | ('T'<<8))
#define _GE         ('G' | ('E'<<8))
#define _LS         ('L' | ('S'<<8))
#define _HI         ('H' | ('I'<<8))
#define _VC         ('V' | ('C'<<8))
#define _VS         ('V' | ('S'<<8))
#define _PL         ('P' | ('L'<<8))
#define _MI         ('M' | ('I'<<8))
#define _CC         ('C' | ('C'<<8))
#define _CS         ('C' | ('S'<<8))
#define _NE         ('N' | ('E'<<8))
#define _EQ         ('E' | ('Q'<<8))

#define _R0         ('R' | ('0'<<8))
#define _R1         ('R' | ('1'<<8))
#define _R2         ('R' | ('2'<<8))
#define _R3         ('R' | ('3'<<8))
#define _R4         ('R' | ('4'<<8))
#define _R5         ('R' | ('5'<<8))
#define _R6         ('R' | ('6'<<8))
#define _R7         ('R' | ('7'<<8))
#define _R8         ('R' | ('8'<<8))
#define _R9         ('R' | ('9'<<8))
#define _R10        ('R' | ('1'<<8) | ('0'<<16))
#define _R11        ('R' | ('1'<<8) | ('1'<<16))
#define _R12        ('R' | ('1'<<8) | ('2'<<16))
#define _R13        ('R' | ('1'<<8) | ('3'<<16))
#define _R14        ('R' | ('1'<<8) | ('4'<<16))
#define _R15        ('R' | ('1'<<8) | ('5'<<16))
#define _SP         ('S' | ('P'<<8))
#define _LR         ('L' | ('R'<<8))
#define _PC         ('P' | ('C'<<8))

#define _DA         ('D' | ('A'<<8))
#define _DB         ('D' | ('B'<<8))
#define _IA         ('I' | ('A'<<8))
#define _IB         ('I' | ('B'<<8))
#define _EA         ('E' | ('A'<<8))
#define _FA         ('F' | ('A'<<8))
#define _ED         ('E' | ('D'<<8))
#define _FD         ('F' | ('D'<<8))



EXPORT_C ASM *assembler(LINE *lines, U32 nlines, void (* cb)(char *, LINE *, U32)) {

  U32 outputsize, line, i;
  char *err, templine[512], *p;
  EVALUATOR *eval;
  ASM *assm;
  void (* callback)(char *, LINE *, U32 type);

  assm = malloc(sizeof(ASM));
  if (!assm)   return NULL;
  memset(assm, 0, sizeof(ASM));

  if (cb)
    callback = cb;
  else
    callback = dummy_callback;

  eval = evaluator_init(NULL);
  if (!eval) {
    callback("Failed to initialise", NULL, ASM_ERROR);
    assembler_kill(assm);
    return NULL;
  }

  // reset stuff
  assm->symbolcount = 0;
  assm->output = NULL;
  assm->errorcount = 0;

  // first pass
  outputsize = 0;

  for (line = 0; line < nlines; line++) {
    err = NULL;

    // make work copy
    strcpy(templine, lines[line].line);
    p = templine;

    if (valid_symbol_start_char(*p)) {
      // lines starting with a letter
      char value[256], *p2;
      // strip trailing spaces
      p2 = p + strlen(p) - 1;
      while (p2[-1] == ' ')  p2--;
      p2[0] = '\0';
      sprintf(value, "%d", outputsize);

      err = define_symbol(assm, eval, p, value, SYMBOL_LABEL);

    } else if (!*p) {
      // empty
    } else if (*p != ' ') {
      // if neither preprocess/command processor/label :
      err = "9Unknown instruction or directive";
    } else {
      // strip leading spaces
      while (*p==' ')  p++;

      // neither comment nor empty line
      if ((*p) && (*p != ';')) {
        U32 n;

        assm->currentline = lines[line].line;
        tidy_instruction(p);
        err = assemble_instruction(assm, eval, p, outputsize, NULL, &n);
        outputsize += n;
      }
    }
  }

  // get memory for the output - alloc some extra for directives
  // which need second pass to return correct size
  assm->totaloutputsize = outputsize + 16*1024;
  assm->output = malloc(assm->totaloutputsize);
  if (!assm->output) {
    evaluator_kill(eval);
    callback("No room", NULL, ASM_ERROR);
    assembler_kill(assm);
    return NULL;
  }

  // second pass
  outputsize = 0;
  for (line = 0; line < nlines; line++) {
    err = NULL;
    // make work copy
    strcpy(templine, lines[line].line);
    p = templine;

    if (*p == ' ') {
      // strip leading spaces
      while (*p==' ')  p++;

      // neither comment nor empty line
      if ((*p) && (*p != ';')) {
        U32 n;

        assm->currentline = lines[line].line;
        tidy_instruction(p);
        err = assemble_instruction(assm, eval, p, outputsize, assm->output+outputsize, &n);
        outputsize += n;
      }
    }
    if (err) {
      if (*err == '0')
        callback(err+1, lines+line, ASM_WARNING);
      else {
        if (assm->errorcount > MAXERRORS) {
          evaluator_kill(eval);
          callback("Too may errors - bailing out", NULL, ASM_ERROR);
          return assm;
        } else {
          strncpy(assm->errors[assm->errorcount].errmsg, err+1, MAXERRORLENGTH-1);
          assm->errors[assm->errorcount].errmsg[MAXERRORLENGTH-1] = '\0';
          assm->errors[assm->errorcount].lineno = lines[line].lineno;
          assm->errors[assm->errorcount].fileno = lines[line].fileno;
        }
        assm->errorcount++;
        callback(err+1, lines+line, ASM_ERROR);
      }
    }
  }

  // check for unused symbols being exported
  for (i = 0; i < assm->symbolcount; i++)
    if (assm->symbols[i].flags & SYMBOL_UNUSED)
      if (assm->symbols[i].flags & SYMBOL_EXPORTED) {
        char errmsg[256];
        sprintf(errmsg, "Unused symbol '%s' is exported", assm->symbols[i].name);
        callback(errmsg, NULL, ASM_ERROR);
      }

  // no need for the evaluator any more
  evaluator_kill(eval);

  assm->totaloutputsize = outputsize;

  if (assm->errorcount) {
    callback("Failed...", NULL, ASM_ERROR);
    return assm;
  }

  sprintf(templine, "Done - %d bytes", outputsize);
  callback(templine, NULL, ASM_INFO);

  return assm;
}


EXPORT_C void assembler_kill(ASM *assm) {

  U32 sym;

  for (sym = 0; sym < assm->symbolcount; sym++) {
    if (assm->symbols[sym].name)   free(assm->symbols[sym].name);
    assm->symbols[sym].name = NULL;
  }
  if (assm->output) {
    free(assm->output);
    assm->output = NULL;
  }
  free(assm);
}


EXPORT_C U32 assembler_errors(ASM *assm, ASMERROR **errors) {
  *errors = assm->errors;
  return assm->errorcount;
}


EXPORT_C void assembler_output(ASM *assm, U32 *size, void **output) {
  *size = assm->totaloutputsize;
  *output = assm->output;
}


EXPORT_C void assembler_symbols(ASM *assm, SYMBOL **s, U32 *ns) {
  *s = assm->symbols;
  *ns = assm->symbolcount;
}

// ------------------------------------------------------------

#define WRITEWORD(p, w) { U32 i, ww; unsigned char *wp;     \
                          ww = w; wp = (unsigned char *)&ww;         \
                          for (i = 0; i < 4; i++)  (p)[i] = wp[i];  }


char *assemble_instruction(ASM *assm, EVALUATOR *eval, char *s, U32 adr, unsigned char *instruc, U32 *bytes) {

  U32 i3, instruction, used, condition, regd, regs1, regs2, regs3;
  char arg1[256], arg2[256], arg3[256], arg4[256], *err, *errmsg;

  errmsg = assm->errmsg;
  instruction = 0;
  *bytes = 4;
  err = NULL;

  i3 = s[0] | (s[1]<<8) | (s[2]<<16);

  switch (i3) {
  case _ADD:
  case _SUB:
  case _ADC:
  case _SBC:
  case _RSB:
  case _RSC:
  case _BIC:
  case _EOR:
  case _ORR:
  case _AND:
    // opcode[COND][S] dest,src1,#const
    // opcode[COND][S] dest,src1,src2[,shift]

    // remove opcode
    s += 3;

    // extract condition
    used = extract_condition(s, &condition);
    instruction |= condition << 28;
    s += used;

    // 'set flags' bit
    if (*s == 'S') {
      s++;
      instruction |= 1<<20;
    }

    if (*s != ' ')   return "9Junk after opcode";

    err = split_in_4(s, arg1, arg2, arg3, arg4);
    if (err)   return err;

    // extract dest register
    regd = extract_register(arg1);
    if (regd == BAD_REGISTER) {
      sprintf(errmsg, "9Bad destination register '%s'", arg1);
      return errmsg;
    }
    instruction |= regd<<12;

    // extract first source register
    regs1 = extract_register(arg2);
    if (regs1 == BAD_REGISTER) {
      sprintf(errmsg, "9Bad source register '%s'", arg2);
      return errmsg;
    }
    instruction |= regs1<<16;

    // third arg is either a register or constant
    regs2 = extract_register(arg3);
    if (regs2 == BAD_REGISTER) {
      // not a register, must be a constant
      S32 constvalue;
      U32 constshift;

      err = constant_8bit(assm, eval, arg3, &constvalue, &constshift);
      if (err)  return err;
      instruction |= constvalue;
      instruction |= constshift<<8;
      instruction |= 1<<25;
      if (*arg4) {
        sprintf(errmsg, "9Junk at end : %s", arg4);
        return errmsg;
      }
    } else {
      // is a register
      instruction |= regs2;
      if (*arg4) {
        U32 shifttype, shiftreg, shiftvalue;
        // there is a shift op on the register
        err = shift(assm, eval, arg4, &shifttype, &shiftreg, &shiftvalue);
        if (err)    return err;
        instruction |= shifttype<<4;
        if (shiftreg == BAD_REGISTER)
          instruction |= shiftvalue<<7;
        else
          instruction |= shiftreg<<8;
      }
    }

    switch (i3) {
    case _AND: instruction |=  0<<21;  break;
    case _EOR: instruction |=  1<<21;  break;
    case _SUB: instruction |=  2<<21;  break;
    case _RSB: instruction |=  3<<21;  break;
    case _ADD: instruction |=  4<<21;  break;
    case _ADC: instruction |=  5<<21;  break;
    case _SBC: instruction |=  6<<21;  break;
    case _RSC: instruction |=  7<<21;  break;
    case _ORR: instruction |= 12<<21;  break;
    case _BIC: instruction |= 14<<21;  break;
    }

    break;
    // ------------------------------------------------------------------
  case _MOV:
  case _MVN:
    // opcode[COND][S] dest,#const
    // opcode[COND][S] dest,src[,shift]

    // remove opcode
    s += 3;

    // extract condition
    used = extract_condition(s, &condition);
    instruction |= condition << 28;
    s += used;

    // 'set flags' bit
    if (*s == 'S') {
      s++;
      instruction |= 1<<20;
    }

    if (*s != ' ') {
      sprintf(errmsg, "9Junk after opcode : %s", s);
      return errmsg;
    }

    err = split_in_4(s, arg1, arg2, arg3, arg4);
    if (err)   return err;

    // extract dest register
    regd = extract_register(arg1);
    if (regd == BAD_REGISTER) {
      sprintf(errmsg, "9Bad destination register '%s'", arg1);
      return errmsg;
    }
    instruction |= regd<<12;

    // second arg is either a register or constant
    regs1 = extract_register(arg2);
    if (regs1 == BAD_REGISTER) {
      // not a register, must be a constant
      S32 constvalue;
      U32 constshift;
      err = constant_8bit(assm, eval, arg2, &constvalue, &constshift);
      if (err)  return err;
      instruction |= constvalue;
      instruction |= constshift<<8;
      instruction |= 1<<25;
      if (*arg3) {
        sprintf(errmsg, "9Junk at end : %s", arg3);
        return errmsg;
      }
    } else {
      // is a register
      instruction |= regs1;
      if (*arg3) {
        U32 shifttype, shiftreg, shiftvalue;
        // there is a shift op on the register
        err = shift(assm, eval, arg3, &shifttype, &shiftreg, &shiftvalue);
        if (err)    return err;
        instruction |= shifttype<<4;
        if (shiftreg == BAD_REGISTER)
          instruction |= shiftvalue<<7;
        else
          instruction |= shiftreg<<8;
      }
      if (*arg4) {
        sprintf(errmsg, "9Junk at end : %s", arg4);
        return errmsg;
      }
    }

    switch (i3) {
    case _MOV: instruction |= 13<<21;  break;
    case _MVN: instruction |= 15<<21;  break;
    }

    break;
    // ------------------------------------------------------------------
  case _CMP:
  case _CMN:
  case _TST:
  case _TEQ:
    // opcode[COND][S] dest,#const
    // opcode[COND][S] dest,src[,shift]

    // remove opcode
    s += 3;

    // extract condition
    used = extract_condition(s, &condition);
    instruction |= condition << 28;
    s += used;

    if (*s != ' ') {
      sprintf(errmsg, "9Junk after opcode : %s", s);
      return errmsg;
    }

    err = split_in_4(s, arg1, arg2, arg3, arg4);
    if (err)   return err;

    // extract dest register
    regd = extract_register(arg1);
    if (regd == BAD_REGISTER) {
      sprintf(errmsg, "9Bad destination register '%s'", arg1);
      return errmsg;
    }
    instruction |= regd<<16;

    // second arg is either a register or constant
    regs1 = extract_register(arg2);
    if (regs1 == BAD_REGISTER) {
      // not a register, must be a constant
      S32 constvalue;
      U32 constshift;

      err = constant_8bit(assm, eval, arg2, &constvalue, &constshift);
      if (err)  return err;
      instruction |= constvalue;
      instruction |= constshift<<8;
      instruction |= 1<<25;
      if (*arg3) {
        sprintf(errmsg, "9Junk at end : %s", arg3);
        return errmsg;
      }
    } else {
      // is a register
      instruction |= regs1;
      if (*arg3) {
        U32 shifttype, shiftreg, shiftvalue;
        // there is a shift op on the register
        err = shift(assm, eval, arg3, &shifttype, &shiftreg, &shiftvalue);
        if (err)    return err;
        instruction |= shifttype<<4;
        if (shiftreg == BAD_REGISTER)
          instruction |= shiftvalue<<7;
        else
          instruction |= shiftreg<<8;
      }
      if (*arg4) {
        sprintf(errmsg, "9Junk at end : %s", arg4);
        return errmsg;
      }
    }

    instruction |= 1<<20;

    switch (i3) {
    case _TST: instruction |= 8<<21;  break;
    case _TEQ: instruction |= 9<<21;  break;
    case _CMP: instruction |= 10<<21;  break;
    case _CMN: instruction |= 11<<21;  break;
    }

    break;
    // ------------------------------------------------------------------
  case _NOP:
    if (s[3]) {
      sprintf(errmsg, "9Junk after opcode : %d", s[3]);
      return errmsg;
    }
    instruction = 0xe1a00000;           // MOV R0,R0
    break;

    // ------------------------------------------------------------------
  case _DCD:
  case _DCW:
  case _DCB:
  case _DCS:
    // skip opcode
    s += 3;
    // skip leading spaces
    while (*s && *s==' ')   s++;

    if (i3 == _DCS) {
      U32 more;

      *bytes = 0;
      if (*s != '\"')  return "9Missing \"";
      more = 1;
      s++;
      do {
        if (!*s)
          more = 0;
        else if (*s=='\"')
          more = 0;
        else if (*s=='\\')
          s++;
        if (more) {
          *bytes += 1;
          if (instruc)  *instruc++ = *s;
        }
        s++;
      } while (more);
      return NULL;
    }

    err = evaluate(assm, eval, s, (S32 *)&instruction);
    if (err)    return err;
    if (i3 == _DCB) {
      if (instruction > 0xff) {
        sprintf(errmsg, "0Outside range for byte : %d", instruction);
        return errmsg;
      }
      *bytes = 1;
    } else if (i3 == _DCW) {
      if (instruction > 0xffff) {
        sprintf(errmsg, "0Outside range for short word : %d", instruction);
        return errmsg;
      }
      *bytes = 2;
    }
    break;
    // ------------------------------------------------------------------
  case _MUL:
  case _MLA:
    // MUL[cond][S] dest,src1,src2
    // MLA[cond][S] dest,src1,src2,src3

    // remove opcode
    s += 3;

    // extract condition
    used = extract_condition(s, &condition);
    instruction |= condition << 28;
    s += used;

    // 'set flags' bit
    if (*s == 'S') {
      s++;
      instruction |= 1<<20;
    }

    if (*s != ' ') {
      sprintf(errmsg, "9Junk after opcode : %s", s);
      return errmsg;
    }

    err = split_in_4(s, arg1, arg2, arg3, arg4);
    if (err)   return err;

    // extract dest register
    regd = extract_register(arg1);
    if (regd == BAD_REGISTER) {
      sprintf(errmsg, "9Bad destination register '%s'", arg1);
      return errmsg;
    }
    instruction |= regd<<16;

    // extract first source register
    regs1 = extract_register(arg2);
    if (regs1 == BAD_REGISTER) {
      sprintf(errmsg, "9Bad 1st source register : %s", arg2);
      return errmsg;
    }
    if (regs1 == regd)    return "9Source and destination register are the same";
    instruction |= regs1<<0;

    // extract second source register
    regs2 = extract_register(arg3);
    if (regs2 == BAD_REGISTER) {
      sprintf(errmsg, "9Bad 2nd source register : %s", arg3);
      return errmsg;
    }
    instruction |= regs2<<8;

    if (i3 == _MLA) {
      regs3 = extract_register(arg4);
      if (regs3 == BAD_REGISTER) {
        sprintf(errmsg, "9Bad 3rd source register : %s", arg4);
        return errmsg;
      }
      instruction |= regs3<<12;
      instruction |= 1<<21;
    } else {
      if (*arg4) {
        sprintf(errmsg, "9Junk after 2nd source register : %s", arg4);
        return errmsg;
      }
    }

    instruction |= 9<<4;

    break;
    // ------------------------------------------------------------------

  case _ADR:
    {
    // ADR[cond] dest,label
    S32 addr;
    U32 op;

    *bytes = 3*4;     // enough to allow VERY long ADR

    // remove opcode
    s += 3;

    // extract condition
    used = extract_condition(s, &condition);
    s += used;

    if (*s != ' ') {
      sprintf(errmsg, "9Junk after opcode : %s", s);
      return errmsg;
    }

    err = split_in_4(s, arg1, arg2, arg3, arg4);
    if (err)   return err;

    // extract dest register
    regd = extract_register(arg1);
    if (regd == BAD_REGISTER) {
      sprintf(errmsg, "9Bad destination register '%s'", arg1);
      return errmsg;
    }
    if (regd == 15)           return "9ADR pc,... is not allowed";

    // skip to label
    s += 2;
    while (*s && *s!=',')   s++;
    if (!*s) {
      sprintf(errmsg, "9Missing second argument : %s", s);
      return errmsg;
    }
    s++;
    // get label
    err = evaluate(assm, eval, s, &addr);
    if (err)   return err;

    // opcode
    addr -= adr+8;
    if (addr < 0) {
      op = 2<<21;
      addr = -addr;
    } else {
      op = 4<<21;
    }

    instruction = (condition<<28) | (op) | (1<<25) | (regd<<12) | (15<<16) | (addr &255);
    if (instruc)   WRITEWORD(instruc+0, instruction)

    if ((addr>>8) &255)
      instruction = (condition<<28) | (op) | (1<<25) | (regd<<12) | (regd<<16) | ((addr>>8)&255);
    else
      instruction = 0xe1a00000;           // MOV R0,R0
    if (instruc)   WRITEWORD(instruc+4, instruction)

    if ((addr>>16) &255)
      instruction = (condition<<28) | (op) | (1<<25) | (regd<<12) | (regd<<16) | ((addr>>16)&255);
    else
      instruction = 0xe1a00000;           // MOV R0,R0
    if (instruc)   WRITEWORD(instruc+8, instruction)

    instruc = NULL;
    }
    break;
    // ------------------------------------------------------------------
  case _LDR:
  case _STR:
    {
    // opcode[cond][B] dest,"[" src[,offset] "]" [!]
    // opcode[cond][B] dest,"[" src "]" [,offset]

    S32 pos1, pos2;

    // strip opcode
    s += 3;

    // extract condition
    used = extract_condition(s, &condition);
    instruction |= condition << 28;
    s += used;

    instruction |= 1<<26;
    if (i3 == _LDR)   instruction |= 1<<20;

    // 'byte' bit
    if (*s == 'B') {
      s++;
      instruction |= 1<<22;
    }

    if (*s != ' ') {
      sprintf(errmsg, "9Junk after opcode : %s", s);
      return errmsg;
    }

    // get destination register
    err = get_until(&s, arg1, ',');
    if (err)    return "9Missing destination register";
    regd = extract_register(arg1);
    if (regd == BAD_REGISTER) {
      sprintf(errmsg, "9Bad register : %s", arg1);
      return errmsg;
    }
    instruction |= regd<<12;

    if (*s != '[') {
      S32 address;
      err = evaluate(assm, eval, s, &address);
      if (err)   return err;
      instruction |= (1<<24) | (15<<16);

      address = address-adr-8;
      if (address < 0)
        address = -address;
      else
        instruction |= 1<<23;
      if (address >= 4096)   return "9Bad offset, must be -4095..4096";
      instruction |= address<<0;
      if (instruc)   WRITEWORD(instruc, instruction)
      return NULL;
    }
    s++;

    // either   src[,offset] "]" [!]   or src "]" [,offset]
    pos1 = strchr(s, ',') - s;
    pos2 = strchr(s, ']') - s;
    if (pos2 <= 0)   return "9Mistake";
    if (pos1 <= 0)   pos1 = 1000;
    if (pos1 < pos2) {
      // src,offset "]" [!]
      instruction |= 1<<24;

      // get source register
      err = get_until(&s, arg1, ',');
      if (err)   return "9Missing source register";
      regs1 = extract_register(arg1);
      if (regs1 == BAD_REGISTER) {
        sprintf(errmsg, "9Bad register : %s", arg1);
        return errmsg;
      }
      instruction |= regs1<<16;

       // what's left:   offset "]" [!]
      if (*s == '#') {
        // offset is a 12 bit constant
        S32 constant;

        s++;
        // find end of constant
        err = get_until(&s, arg1, ']');
        if (err)   return "9Missing ]";
        err = evaluate(assm, eval, arg1, &constant);
        if (err)   return err;
        if (constant < 0)
          constant = -constant;
        else
          instruction |= 1<<23;
        if (constant >= 4096)   return "9Bad constant, must be -4095..4095";
        instruction |= constant<<0;

      } else {
        // offset is a register
        U32 negative;

        instruction |= 1<<25;

        // sign
        if (*s == '-') {
          s++;
          negative = 1;
        } else {
          instruction |= 1<<23;
          negative = 0;
        }

        // either reg"]"                          or
        //        reg,shifttype #const"]"         or
        //        reg,shifttype reg"]"
        pos1 = strchr(s, ',') - s;
        pos2 = strchr(s, ']') - s;
        if (pos1 < 0) {
          // reg"]"
          err = get_until(&s, arg1, ']');
          if (err)    return err;
          regs2 = extract_register(arg1);
          if (regs2 == BAD_REGISTER) {
            sprintf(errmsg, "9Bad register : %s", arg1);
            return errmsg;
          }
          instruction |= regs2<<0;

        } else {
        // either   reg,shifttype #const"]"         or
        //          reg,shifttype reg"]"
          U32 shifttype, shiftreg, shiftvalue;

          err = get_until(&s, arg1, ',');
          if (err)   return err;
          regs2 = extract_register(arg1);
          if (regs2 == BAD_REGISTER) {
            sprintf(errmsg, "9Bad register : %s", arg1);
            return errmsg;
          }
          instruction |= regs2<<0;

          err = get_until(&s, arg1, ']');
          if (err)   return err;
          err = shift(assm, eval, arg1, &shifttype, &shiftreg, &shiftvalue);
          if (err)   return err;
          instruction |= shifttype<<4;
          if (shiftreg == BAD_REGISTER)
            instruction |= shiftvalue<<7;
          else
            instruction |= shiftreg<<8;
        }
      }
      if (*s == '!') {
        s++;
        instruction |= 1<<21;
      }
      if (*s) {
        sprintf(errmsg, "9Junk at end : %s", s);
        return errmsg;
      }

    } else {
      // src "]" [,offset]

      err = get_until(&s, arg1, ']');
      if (err)   return err;
      regs1 = extract_register(arg1);
      if (regs1 == BAD_REGISTER) {
        sprintf(errmsg, "9Bad register : %s", arg1);
        return errmsg;
      }
      instruction |= regs1<<16;

      if (*s == ',') {
        // skip the comma and the trailing spaces
        s++;
        if (*s == '#') {
          // offset is a constant
          S32 constant;
          s++;
          err = evaluate(assm, eval, s, &constant);
          if (constant < 0)
            constant = -constant;
          else
            instruction |= 1<<23;
          if (constant >= 4096)   return "9Bad constant, must be -4095..4095";
          instruction |= constant<<0;

        } else {
          // offset is register + optional shift
          instruction |= 1<<25;
          if (*s == '-')
            s++;
          else
            instruction |= 1<<23;

          if (strchr(s, ',') == 0) {
            // no shift
            regs2 = extract_register(s);
            if (regs2 == BAD_REGISTER) {
              sprintf(errmsg, "9Bad register : %s", s);
              return errmsg;
            }
            instruction |= regs2<<0;

          } else {
            // has shift
            U32 shifttype, shiftreg, shiftvalue;

            err = get_until(&s, arg1, ',');
            if (err)   return err;
            regs2 = extract_register(arg1);
            if (regs2 == BAD_REGISTER) {
              sprintf(errmsg, "9Bad register : %s", arg1);
              return errmsg;
            }
            instruction |= regs2<<0;

            err = shift(assm, eval, s, &shifttype, &shiftreg, &shiftvalue);
            if (err)   return err;
            instruction |= shifttype<<4;
            if (shiftreg != BAD_REGISTER)
              return "9Bad shift, shift bitcount mustn't be a register";
            instruction |= shiftvalue<<7;
          }
        }

      } else if (*s) {
        sprintf(errmsg, "9Junk after ] : %s", s);
        return errmsg;
      } else
        // no offset, but just to make things a little prettier, set 'positive' bit
        instruction |= 1<<23;
    }

    }
    break;
    // ------------------------------------------------------------------
  case _LDM:
  case _STM:
    {
    // opcode[cond]heaptype source[!],{reglist}
    U32 stacktype, done;

    instruction |= 1<<27;

    // strip opcode
    s += 3;

    // extract condition
    used = extract_condition(s, &condition);
    instruction |= condition << 28;
    s += used;

    stacktype = s[0] | (s[1]<<8);
    if (i3 == _LDM) {
      instruction |= 1<<20;
      switch (stacktype) {
      case _FA:
      case _DA:
        instruction |= 0<<23;
        break;
      case _FD:
      case _IA:
        instruction |= 1<<23;
        break;
      case _DB:
      case _EA:
        instruction |= 2<<23;
        break;
      case _ED:
      case _IB:
        instruction |= 3<<23;
        break;
      default:
        sprintf(errmsg, "9Unsupported LDM/STM type : %c%c", s[0], s[1]);
        return errmsg;
        break;
      }
    } else {
      switch (stacktype) {
      case _DA:
      case _ED:
        instruction |= 0<<23;
        break;
      case _IA:
      case _EA:
        instruction |= 1<<23;
        break;
      case _DB:
      case _FD:
        instruction |= 2<<23;
        break;
      case _IB:
      case _FA:
        instruction |= 3<<23;
        break;
      default:
        sprintf(errmsg, "9Unsupported LDM/STM type : %c%c", s[0], s[1]);
        return errmsg;
        break;
      }
    }
    // strip type
    s += 2;

    if (*s != ' ') {
      sprintf(errmsg, "9Junk after end : %s", s);
      return errmsg;
    }

    // extract source register
    err = get_until(&s, arg1, ',');
    if (err)  return err;

    if (arg1[strlen(arg1)-1] == '!') {
      instruction |= 1<<21;
      arg1[strlen(arg1)-1] = '\0';
    }
    regs1 = extract_register(arg1);
    if (regs1 == BAD_REGISTER) {
      sprintf(errmsg, "9Bad register : %s", arg1);
      return errmsg;
    }
    instruction |= regs1<<16;

    // get to start of register list
    if (*s != '{')            return "9Missing {";
    if (!strchr(s, '}'))      return "9Missing }";
    s++;


    done = 0;
    while (*s && !done) {
      char *s1, *s2, *s3, *err1, *err2, *err3;
      S32 p1, p2, p3;
      s1 = s;
      s2 = s;
      s3 = s;
      err1 = get_until(&s1, arg1, '-');
      err2 = get_until(&s2, arg2, ',');
      err3 = get_until(&s3, arg3, '}');
      p1 = s1 - s;
      p2 = s2 - s;
      p3 = s3 - s;
      if (err1)  p1 = 1000;
      if (err2)  p2 = 1000;

      if (err1 && err2) {
        // single register
        s = s3;
        done = 1;
        regs1 = extract_register(arg3);
        if (regs1 == BAD_REGISTER) {
          sprintf(errmsg, "9Bad register : %s", arg3);
          return errmsg;
        }
        if (instruction & (1<<regs1))   err = "0Dublicate register in list";
        instruction |= 1<<regs1;
      } else if (p1 < p2) {
        // list
        regs1 = extract_register(arg1);
        if (regs1 == BAD_REGISTER) {
          sprintf(errmsg, "9Bad register : %s", arg1);
          return errmsg;
        }
        s = s1;
        if (err2) {
          // no , so it is the last element in the list
          err = get_until(&s, arg2, '}');
          done = 1;
        } else {
          // there is a ,
          err = get_until(&s, arg2, ',');
        }
        if (err)                        return "9Bad register list";
        regs2 = extract_register(arg2);
        if (regs2 == BAD_REGISTER) {
          sprintf(errmsg, "9Bad register : %s", arg2);
          return errmsg;
        }
        if (regs2 < regs1)              return "9Bad register list";
        for ( ; regs1 <= regs2; regs1++) {
          if (instruction & (1<<regs1)) err = "0Dublicate register in list";
          instruction |= 1<<regs1;
        }

      } else {
        // single register
        s = s2;
        regs1 = extract_register(arg2);
        if (regs1 == BAD_REGISTER) {
          sprintf(errmsg, "9Bad register : %s", arg2);
          return errmsg;
        }
        if (instruction & (1<<regs1))   err = "0Dublicate register in list";
        instruction |= 1<<regs1;
      }
    }

    if (*s == '^') {
      // HBP - only in 26 bit mode?/
    } else if (*s) {
      sprintf(errmsg, "9Junk at end : %s", s);
      return errmsg;
    }


    }
    break;
    // ------------------------------------------------------------------
  default:
    if (*s == 'B') {
      U32 bl;
      S32 address;
      // B or BL

      bl = 0;
      // skip opcode
      s++;
      // extract condition and B/BL
      if (*s == 'L') {
        if (extract_condition(s+1, &condition) == 2) {
          bl = 1;
          used = 3;
        } else if (s[1]==' ') {
          bl = 1;
          used = 1;
        } else
          used = extract_condition(s, &condition)+1;
      } else {
        used = extract_condition(s, &condition);
      }
      instruction |= condition << 28;
      // skip condition
      s += used;

      if (bl)
        instruction |= 11<<24;
      else
        instruction |= 10<<24;
      err = evaluate(assm, eval, s, &address);
      if (err)   return err;
      instruction |= ((address-adr-8)>>2) & 0x00ffffff;

    } else if (strncmp(s, "CONST ", 6) == 0) {
      // define constant
      char *p, *endofname;

      // doesn't generate output
      *bytes = 0;

      s += 6;
      // strip spaces
      while (*s == ' ')  s++;

      p = strchr(s, '=');
      if (!p)    return "9Missing = in CONST";

      endofname = p-1;
      while (*endofname == ' ')  *endofname-- = '\0';
      *p++ = '\0';

      if (*s && *p) {
        if (strlen(s) > MAXVARNAMELENGTH)   return "9Bad name length in CONST";
        if (!instruc)  return define_symbol(assm, eval, s, p, SYMBOL_CONSTANT);
        return NULL;
      }
      return "9Bad CONST statement";

    } else if (strncmp(s, "EXPORT ", 7) == 0) {
      // mark symbols for export
      char *symbollist[256], **list;
      U32 n, matched;

      // doesn't generate output
      *bytes = 0;

      err = split_symbol_list(assm, s+7, symbollist);
      if (err)   return err;

      list = symbollist;
      while (*list) {
        matched = 0;
        for (n = 0; n < assm->symbolcount; n++)
          if (strcmp(assm->symbols[n].name, *list) == 0) {
            assm->symbols[n].flags |= SYMBOL_EXPORTED;
            matched++;
          }
        if (matched == 0) {
          err = define_symbol(assm, eval, *list, "0", SYMBOL_UNUSED | SYMBOL_EXPORTED);
          if (err)   return err;
        }

        list++;
      }
#ifdef SUPPORT_IMPORT
    } else if (strncmp(s, "IMPORTB ", 8) == 0) {
      // mark create list of symbols for import as 'B xxxx'
      char *symbollist[128], **list;

      // so far, no extra output bytes
      *bytes = 0;

      err = split_symbol_list(assm, s+8, symbollist);
      if (err)   return err;

      list = symbollist;
      while (*list) {
        if (!instruc) {
          char value[128];
          sprintf(value, "%d", adr+*bytes);
          err = define_symbol(assm, eval, *list, value, SYMBOL_BRANCHLABEL | SYMBOL_IMPORTED);
        }
        if (instruc)   WRITEWORD(instruc + *bytes, 0xe1a0f00e)     // MOV PC,R14
        *bytes += 4;
        list++;
      }
      instruc = NULL;

    } else if (strncmp(s, "IMPORT ", 7) == 0) {
      // mark create list of symbols for import as 'EQUD xxxx'
      char *symbollist[128], **list;

      // so far, no extra output bytes
      *bytes = 0;

      err = split_symbol_list(assm, s+7, symbollist);
      if (err)   return err;

      list = symbollist;
      while (*list) {
        if (!instruc) {
          char value[128];
          sprintf(value, "%d", adr+*bytes);
          err = define_symbol(assm, eval, *list, value, SYMBOL_CONSTANT | SYMBOL_IMPORTED);
        }
        if (instruc)   WRITEWORD(instruc + *bytes, 0)
        *bytes += 4;
        list++;
      }
      instruc = NULL;
#endif

    } else if (strncmp(s, "INCLUDEFILE ", 12) == 0) {
      char *p;
      FILE *fh;
      long filesize;

      p = s+12;
      while (*p == ' ')   p++;
      if (!*s)    return "9Not filename specified";
      // open file
      fh = fopen(p, "rb");
      if (!fh) {
        sprintf(errmsg, "9Failed to open '%s'", p);
        return errmsg;
      }
      // get file size
      fseek(fh, 0, SEEK_END);
      filesize = ftell(fh);
      *bytes = (U32)filesize;
      if (adr+*bytes > assm->totaloutputsize) {
        fclose(fh);
        return "9Failed to load datafile";
      }
      if (instruc) {
        // back to start, then read the entire file
        fseek(fh, 0, SEEK_SET);
        fread(instruc, 1, (int)filesize, fh);
      }
      fclose(fh);

    } else if (strncmp(s, "SYMBOLLIST ", 11) == 0) {
      U32 lbls, cnsts, all, n, flags, sym;
      char *p;
      S32 prog[16];

      p = strstr(assm->currentline, "SYMBOLLIST") + 11;
      n = sscanf(p, "%x %x %x %x %x %x %x %x %x %x %x %x %x %x %x %x %x",
                 &flags, prog+0, prog+1, prog+2, prog+3,
                         prog+4, prog+5, prog+6, prog+7,
                         prog+8, prog+9, prog+10, prog+11,
                         prog+12, prog+13, prog+14, prog+15);
      n -= 1;
      if (n < 1)              return "9Bad SYMBOLLIST (too few arguments)";
      if ((flags &3) == 0)    return "9Bad SYMBOLLIST (bad flags)";
      lbls = cnsts = all = 0;
      if (flags & 1)   lbls = 1;
      if (flags & 2)   cnsts = 1;
      if (flags & 4)   all = 1;
      *bytes = 0;
      for (sym = 0; sym < assm->symbolcount; sym++) {
        SYMBOL *symbol;
        symbol = &assm->symbols[sym];
        if (  (all || (symbol->flags & SYMBOL_EXPORTED)) &&
              ( (lbls && (symbol->flags & SYMBOL_LABEL))    ||
                (cnsts && (symbol->flags & SYMBOL_CONSTANT)) )  ) {
          U32 np;
          for (np = 0; np < n; np++) {
            U32 opcode, left, nbytes;
            S32 arg;

            nbytes = 0;
            opcode = prog[np] & 15;     // opcode in bits 0..3
            arg = prog[np]>>8;          // arg in bits 8..31

            left = assm->totaloutputsize - adr;
            switch (opcode) {
            case 0:           // write bytes
              // arg: bits  8..23  length
              //      bits  0..7   value to write
              nbytes = arg>>8;
              if (instruc) {
                U32 i;
                if (left < nbytes)   return "9No room for SYMBOLLIST";
                for (i = 0; i < nbytes; i++)   *instruc++ = arg &255;
              }
              break;
            case 1:           // write symbolvalue
              // arg: bit   0      0 -> write 32 bits
              //                   1 -> write 16 bits
              //      bits  8..23  signed value to add to symbolvalue
              if (arg &1)
                nbytes = 2;
              else
                nbytes = 4;
              if (instruc) {
                S32 v;
                if (left < nbytes)   return "9No room for SYMBOLLIST";
                v = symbol->value + (arg>>16);
                *instruc++ = v &255;
                *instruc++ = (v>>8) &255;
                if (nbytes == 4) {
                  *instruc++ = (v>>16) &255;
                  *instruc++ = (v>>24) &255;
                }
              }
              break;
            case 2:           // write name
              // arg: bits  0..1        1: lowercase
              //                        2: uppercase
              //                        otherwise no change
              nbytes = strlen(symbol->name);
              if (instruc) {
                U32 i;

                if (left < nbytes)   return "9No room for SYMBOLLIST";
                for (i = 0; i < nbytes; i++) {
                  switch (arg &3) {
                  case 1:     // lowercase
                    *instruc++ = tolower(symbol->name[i]);
                    break;
                  case 2:     // uppercase
                    *instruc++ = toupper(symbol->name[i]);
                    break;
                  default:
                    *instruc++ = symbol->name[i];
                    break;
                  }
                }
              }
              break;
            case 3:           // align
              // arg: bits 0..7    multiple to align to
              {
                U32 boundary;

                boundary = arg &255;
                if (boundary == 0)  boundary = 4;
                if ((adr + *bytes) % boundary)
                  nbytes = boundary-((adr + *bytes) % boundary);
                else
                  nbytes = 0;
                if (instruc) {
                  U32 i;

                  if (left < nbytes)   return "9No room for SYMBOLLIST";
                  for (i = 0; i < nbytes; i++)  *instruc++ = '\0';
                }
              }
              break;
            case 4:           // write name length
              // arg: bits  0..1   no. of bytes to write (0 to write 4 bytes)
              //      bits  8..23  signed value to add to namelength
              nbytes = arg &3;
              if (nbytes == 0)   nbytes = 4;
              if (instruc) {
                S32 nl;
                if (left < nbytes)   return "9No room for SYMBOLLIST";
                nl = strlen(symbol->name) + (arg>>8);
                *instruc++ = nl &255;
                if (nbytes > 1)   *instruc++ = (nl>>8) &255;
                if (nbytes > 2)   *instruc++ = (nl>>16) &255;
                if (nbytes > 3)   *instruc++ = (nl>>24) &255;
              }
              break;
            }
            *bytes += nbytes;
          }
        }
      }
      instruc = NULL;

    } else if (strncmp(s, "ALIGN", 5) == 0) {
      // ALIGN [boundary]
      S32 boundary;
      // strip opcode and trailing spaces
      s += 5;
      while (*s && *s==' ')  s++;
      if (*s) {
        err = evaluate(assm, eval,s, &boundary);
        if (err)   return err;
        if (boundary < 0 || boundary > 4) {
          sprintf(errmsg, "9Bad ALIGN boundary : %d", boundary);
          return errmsg;
        }

      } else // default boundary is 4 (word-align)
        boundary = 4;

      instruction = 0;
      if (adr % boundary)
        *bytes = boundary-(adr % boundary);
      else
        *bytes = 0;

      // write NULL bytes
      if (instruc) {
        U32 i;
        for (i = 0; i < *bytes; i++)   instruc[i] = '\0';
      }

    } else {
      sprintf(errmsg, "9Unknown opcode : %s", s);
      return errmsg;
    }
    break;
  }

  if (instruc)   WRITEWORD(instruc+0, instruction)

  return err;
}

// ----------------------------------------------------------------------------


char *shift(ASM *assm, EVALUATOR *eval, char *c, U32 *type, U32 *reg, U32 *shift) {

  char *err;

  if (strncmp(c, "RRX", 3) == 0) {
    *type = 6;
    *reg = BAD_REGISTER;
    *shift = 0;
    return NULL;
  }
  if (strchr(c, '#')) {
    *reg = -1;
    err = evaluate(assm, eval, strchr(c, '#')+1, (S32 *)shift);
    if (err)   return err;
    if (*shift >= 32)   return "4Bad shift";
  } else {
    *shift = BAD_REGISTER;
    *reg = extract_register(c+3);
    if (*reg == -1)  return "9Bad shift register";
  }
  if (strncmp(c, "ROR", 3) == 0)
    *type = 6;
  else if (strncmp(c, "ASR", 3) == 0)
    *type = 4;
  else if (strncmp(c, "LSR", 3) == 0)
    *type = 2;
  else if ((strncmp(c, "ASL", 3) == 0) || (strncmp(c, "LSL", 3) == 0))
    *type = 0;
  else
    return "9Bad shifttype";

  if (*reg != BAD_REGISTER)    *type = *type + 1;
  return NULL;
}


char *constant_8bit(ASM *assm, EVALUATOR *eval, char *s, S32 *value, U32 *shift) {

  U32 c, sh, shifted;
  char *err;

  if (*s != '#')    return "9Bad constant";

  err = evaluate(assm, eval, s+1, (int *)&c);
  if (err)   return err;
  if (c < 256) {
    *shift = 0;
    *value = c;
    return NULL;
  }
  sh = 0;
  do {
    sh += 2;
    shifted = (c>>sh) | (c<<(32-sh));
  } while (sh != 32 && shifted >= 256);
  if (sh == 32)    return "4Shiftvalue out of range";
  *shift = (32-sh)/2;
  *value = shifted;
  return NULL;
}


char *define_symbol(ASM *assm, EVALUATOR *eval, char *name, char *value, U32 flags) {
// define a symbol, or set the symbol value if the symbol was previously defined
// using the SYMBOL_UNUSED flags
  char *err;
  U32 i;
  S32 unused;

  if (assm->symbolcount >= MAXSYMBOLS)      return "9Too many symbols";

  // validate name
  if (!name[0])                             return "9Illegal symbol (empty name)";
  if (!valid_symbol_start_char(name[0]))    return "9Illegal symbol (bad name)";
  for (i = 1; name[i]; i++)
    if (!valid_symbol_char(name[i]))        return "9Illegal symbol (bad name)";

  unused = -1;
  for (i = 0; (i < assm->symbolcount) && (unused == -1); i++)
    if (strcmp(name, assm->symbols[i].name) == 0) {
      if (assm->symbols[i].flags & SYMBOL_UNUSED)
        unused = i;
      else
        return "4Dublicate symbol definition";
    }

  if (unused != -1) {
    // no longer unused
    assm->symbols[unused].flags &= ~SYMBOL_UNUSED;
    // set the new flags
    assm->symbols[unused].flags |= flags & 0xffff;
    // evaluate new value
    err = evaluate(assm, eval, value, &assm->symbols[unused].value);
    if (err)   return err;
    // update the evaluator's copy
    set_variable_value(eval, assm->symbols[unused].name, assm->symbols[unused].value);
    return NULL;
  }

  assm->symbols[assm->symbolcount].name = malloc(strlen(name)+1);
  if (!assm->symbols[assm->symbolcount].name)           return "9No room";

  strcpy(assm->symbols[assm->symbolcount].name, name);
  assm->symbols[assm->symbolcount].flags = flags;
  err = evaluate(assm, eval, value, &assm->symbols[assm->symbolcount].value);
  if (err)   return err;

  if (create_variable(eval, name))          return "9Failed to create variable";

  if (set_variable_value(eval, name, assm->symbols[assm->symbolcount].value))
                                            return "9Failed to create variable";

  assm->symbolcount++;

  return NULL;
}


char *evaluate(ASM *assm, EVALUATOR *eval, char *s, S32 *value) {

  S32 err;

  err = evaluate_expression(eval, s, (S32 *)value);
  if (err < 0) {
    // possibly not serious error
    sprintf(assm->errmsg, "4Failed to evaluate '%s'", s);
    return assm->errmsg;
  } else if (err > 0) {
    // definately serious error
    sprintf(assm->errmsg, "9Failed to evaluate '%s'", s);
    return assm->errmsg;
  }

  return NULL;
}


char *split_symbol_list(ASM *assm, char *p, char **list) {

  do {
    // strip spaces
    while (*p == ' ')  p++;
    if (!valid_symbol_start_char(*p)) {
      sprintf(assm->errmsg, "9Illegal symbol or symbollist '%s'", p);
      return assm->errmsg;
    }

    *list++ = p;

    while ((*p) && (valid_symbol_char(*p)))    p++;
    if ((*p) && (*p != ',') && (*p != ' '))  return "9Illegal symbol list seperator";
    *p = '\0';

    p++;
    // strip comma (if present)
    if (*p == ',')  p++;
    // strip spaces
    while (*p == ' ')  p++;
  } while (*p);

  *list = NULL;
  return NULL;
}

// ----------------------------------------------------------------------------

char *get_until(char **src, char *out, char split) {

  S32 p;

  p = strchr(*src, split) - *src;
  if (p <= 0)   return "9Missing split character";
  strncpy(out, *src, p);
  out[p] = '\0';
  *src += p+1;

  return NULL;
}


void tidy_instruction(char *p) {

  char *rd, *wr;
  U32 quote;

  // remove comment
  if (strchr(p, ';'))   *(strchr(p, ';')) = '\0';

  rd = wr = p;
  // uppercase until first space
  while ((*rd) && (*rd != ' '))   *wr++ = toupper(*rd++);
  // copy first space
  if (*rd == ' ')  *wr++ = *rd++;
  // ignore all remaining spaces
  quote = 0;
  while (*rd) {
    if ((*rd == '\"') && (rd[-1] != '\\'))   quote = !quote;
    if ((*rd != ' ') || (quote)) *wr++ = *rd;
    rd++;
  }
  // remove trailing spaces
  while (wr[-1] == ' ')  wr--;

  wr[0] = '\0';
}


U32 valid_symbol_char(char c) {
  if (c >= 'a' && c <= 'z')  return 1;
  if (c >= 'A' && c <= 'Z')  return 1;
  if (c >= '0' && c <= '9')  return 1;
  if (c == '_')   return 1;
  return 0;
}


U32 valid_symbol_start_char(char c) {
  if (c >= 'a' && c <= 'z')  return 1;
  if (c >= 'A' && c <= 'Z')  return 1;
  if (c == '_')   return 1;
  return 0;
}


U32 extract_register(char *r) {

  int reg, used;

  // skip leading spaces
  while (*r == ' ')   r++;

  reg = toupper(r[0]);
  if (r[1])   reg |= toupper(r[1])<<8;
  if (r[2])   reg |= toupper(r[2])<<16;
  used = 2;

  switch (reg) {
  case _R0:   reg = 0;  used = 2;  break;
  case _R1:   reg = 1;  used = 2;  break;
  case _R2:   reg = 2;  used = 2;  break;
  case _R3:   reg = 3;  used = 2;  break;
  case _R4:   reg = 4;  used = 2;  break;
  case _R5:   reg = 5;  used = 2;  break;
  case _R6:   reg = 6;  used = 2;  break;
  case _R7:   reg = 7;  used = 2;  break;
  case _R8:   reg = 8;  used = 2;  break;
  case _R9:   reg = 9;  used = 2;  break;
  case _R10:  reg = 10; used = 3;  break;
  case _R11:  reg = 11; used = 3;  break;
  case _R12:  reg = 12; used = 3;  break;
  case _SP:   reg = 13; used = 2;  break;
  case _R13:  reg = 13; used = 3;  break;
  case _LR:   reg = 14; used = 2;  break;
  case _R14:  reg = 14; used = 3;  break;
  case _PC:   reg = 15; used = 2;  break;
  case _R15:  reg = 15; used = 3;  break;
  default:
    reg = BAD_REGISTER;
    break;
  }
  if (reg == BAD_REGISTER)    return BAD_REGISTER;
  if (r[used])                return BAD_REGISTER;

  return reg;
}


U32 extract_condition(char *c, U32 *condition) {

  U32 used, s;

  s = toupper(c[0]) | (toupper(c[1])<<8);
  used = 2;

  switch (s) {
  case _NV:  *condition = 15 ; break;
  case _AL:  *condition = 14 ; break;
  case _LE:  *condition = 13 ; break;
  case _GT:  *condition = 12 ; break;
  case _LT:  *condition = 11 ; break;
  case _GE:  *condition = 10 ; break;
  case _LS:  *condition = 9  ; break;
  case _HI:  *condition = 8  ; break;
  case _VC:  *condition = 7  ; break;
  case _VS:  *condition = 6  ; break;
  case _PL:  *condition = 5  ; break;
  case _MI:  *condition = 4  ; break;
  case _CC:  *condition = 3  ; break;
  case _CS:  *condition = 2  ; break;
  case _NE:  *condition = 1  ; break;
  case _EQ:  *condition = 0  ; break;
  default:
    *condition = 14;
    used = 0;
    break;
  }
  return used;
}


char *split_in_4(char *s, char *arg1, char *arg2, char *arg3, char *arg4) {

  *arg1 = *arg2 = *arg3 = *arg4 = '\0';

  while (*s && *s!=',')   *arg1++ = *s++;
  *arg1 = '\0';
  if (*s)  s++;

  while (*s && *s!=',')   *arg2++ = *s++;
  *arg2 = '\0';
  if (*s)  s++;

  if (strchr(s, ',')) {
    while (*s && *s!=',')   *arg3++ = *s++;
    *arg3 = '\0';
    s++;
    strcpy(arg4, s);
  } else {
    strcpy(arg3, s);
  }

  return NULL;
}


void dummy_callback(char *err, LINE *line, U32 type) {
// dummy function in case the client doesn't provide a callback function
// can be use to send errors to stderr or to a file or wherever
}
