#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <math.h>
#include <ctype.h>

#include "proto.h"
#include "evaluate.h"


#define CONSTANT_e            2.71828182846
#define CONSTANT_PI           3.14159265359


struct VARIABLE {
  char name[MAXVARNAMELENGTH+1];
  S32 value;
};

struct EVALUATOR {
  VARIABLE vars[MAXVARS];
  void (* callback)(char *);
};


static double evaluate(EVALUATOR *eval, char *string, S32 *error);
static S32 find_variable(EVALUATOR *eval, char *name);
static void dummy_callback(char *err);


EXPORT_C EVALUATOR *evaluator_init(void (* cb)(char *)) {

  S32 i;
  EVALUATOR *eval;

  eval = malloc(sizeof(EVALUATOR));
  if (!eval)   return NULL;

  for (i = 0; i < MAXVARS; i++)    eval->vars[i].name[0] = '\0';
  if (cb)
    eval->callback = cb;
  else
    eval->callback = dummy_callback;

  return eval;
}


EXPORT_C void evaluator_kill(EVALUATOR *eval) {
  free(eval);
}


EXPORT_C S32 evaluate_expression(EVALUATOR *eval, char *src, S32 *value) {
// do a bit of preprocessing and then evaluate the expression

  S32 read, write, i, parentheses, error;
  double res;
  char expr[512];

  strcpy(expr, src);

  // check for unmatched parentheses
  parentheses = 0;
  i = 0;
  while (expr[i] >= 32) {
    if (expr[i] == '(') parentheses++;
    if (expr[i] == ')') parentheses--;
    if (parentheses < 0)  return EVALUATE_UNMATCHED_BRACKET;
    i++;
  }
  if (parentheses != 0)  return EVALUATE_UNMATCHED_BRACKET;

  // remove spaces
  read = write = 0;
  while (expr[read] >= 32) {
    if (expr[read] > 32) {
      expr[write] = expr[read];
      write++;
    }
    read++;
  }

  // terminate expression
  expr[write] = 0;

  // evaluate
  error = 0;
  res = evaluate(eval, expr, &error);
  if (error)    return error;
  *value = (S32)res;
  return EVALUATE_OK;
}


EXPORT_C S32 variable_exists(EVALUATOR *eval, char *var) {

  if (find_variable(eval, var) >= 0)  return EVALUATE_MISSING_VARIABLE;

  return 0;
}


EXPORT_C S32 create_variable(EVALUATOR *eval, char *var) {

  S32 vi;
  char errmsg[256];

  if ((strlen(var) > MAXVARNAMELENGTH) || (!*var)) {
    eval->callback("Illegal varname\n");
    return EVALUATE_ILLEGAL_VARIABLENAME;
  }

  if (variable_exists(eval, var)) {
    sprintf(errmsg, "Variable '%s' already exists\n", var);
    eval->callback(errmsg);
    return EVALUATE_VARIABLE_ALREADY_EXISTS;
  }

  for (vi = 0; vi < MAXVARS; vi++)
    if (eval->vars[vi].name[0] == '\0') {
      strcpy(eval->vars[vi].name, var);
      eval->vars[vi].value = 0;
      return EVALUATE_OK;
    }

  sprintf(errmsg, "Failed to create variable '%s'\n", var);
  eval->callback(errmsg);
  return EVALUATE_TOO_MANY_VARIABLES;
}


EXPORT_C S32 destroy_variable(EVALUATOR *eval, char *var) {

  S32 vi;

  vi = find_variable(eval, var);
  if (vi < 0) {
    char errmsg[256];
    sprintf(errmsg, "Variable '%s' not found\n", var);
    eval->callback(errmsg);
    return EVALUATE_MISSING_VARIABLE;
  }
  eval->vars[vi].name[0] = '\0';

  return EVALUATE_OK;
}


EXPORT_C S32 set_variable_value(EVALUATOR *eval, char *var, S32 value) {

  S32 vi;

  vi = find_variable(eval, var);
  if (vi < 0) {
    char errmsg[256];
    sprintf(errmsg, "Variable '%s' not found\n", var);
    eval->callback(errmsg);
    return EVALUATE_MISSING_VARIABLE;
  }
  eval->vars[vi].value = value;

  return EVALUATE_OK;
}


EXPORT_C S32 read_variable_value(EVALUATOR *eval, char *var, S32 *value) {

  S32 vi;

  vi = find_variable(eval, var);
  if (vi < 0) {
    char errmsg[256];
    sprintf(errmsg, "Variable '%s' not found\n", var);
    eval->callback(errmsg);
    return EVALUATE_MISSING_VARIABLE;
  }
  *value = eval->vars[vi].value;

  return EVALUATE_OK;
}


// ----------------------------------------------------

S32 find_variable(EVALUATOR *eval, char *name) {

  S32 vi;

  for (vi = 0; vi < MAXVARS; vi++)
    if (strcmp(eval->vars[vi].name, name) == 0)   return vi;

  return EVALUATE_MISSING_VARIABLE;
}

/*
 *   + - * /
 *   < > == !> <= >=       1+2<3*4  =  (1+2)<(3*4)    cmps return 1 (true) or 0 (false)
 *   && || & |
 * numbers (integers or floats)
 * PI and e
 * functions
 *   trig
 *      SIN(expr)
 *      COS(expr)
 *      TAN(expr)
 *   math
 *      LOG(expr)
 *      EXP(expr)
 *      SQR(expr)
 *      ABS(expr)
 *      RND(expr)
 *
 */


#define FUNC_SIN (('s'<<16) | ('i'<<8) | ('n'))
#define FUNC_COS (('c'<<16) | ('o'<<8) | ('s'))
#define FUNC_TAN (('t'<<16) | ('a'<<8) | ('n'))
#define FUNC_EXP (('e'<<16) | ('x'<<8) | ('p'))
#define FUNC_LOG (('l'<<16) | ('o'<<8) | ('g'))
#define FUNC_ABS (('a'<<16) | ('b'<<8) | ('s'))
#define FUNC_SQR (('s'<<16) | ('q'<<8) | ('r'))
#define FUNC_RND (('r'<<16) | ('n'<<8) | ('d'))


double evaluate(EVALUATOR *eval, char *s, S32 *error) {

  S32 more, parentheses, i, unnecessary, length, func, varno;
  double value1, value2;
  float value4;

  // find the length of the string
  length = strlen(s);

  if (length == 0)    return 0;

  // remove unncessary parentheses
  do {
    more = 0;
    if ((s[0] == '(') && (s[length-1] == ')')) {
      parentheses = 0;
      unnecessary = 1;
      i = 0;
      while (i < length-1) {
        if (s[i] == '(') parentheses++;
        if (s[i] == ')') parentheses--;
        if (parentheses == 0) unnecessary = 0;
        i++;
      }
      if (unnecessary) {
        for (i = 0; i < length-2; i++)  s[i] = s[i+1];
        length -= 2;
        s[length] = 0;
        more = 1;
      }
    }
  } while (more);

  // scan for comparison and certain ops outside parentheses
  parentheses = 0;
  for (i = 0; i < length-1; i++) {
    if (s[i] == '(') parentheses++;
    if (s[i] == ')') parentheses--;
    if ( (parentheses == 0) &&
         ((s[i] == '=') || (s[i] == '<') || (s[i] == '>') || (s[i] == '!') ||
          (s[i] == '&') || (s[i] == '|'))) {
      char *p;
      S32 cmp;

#define CMP_EQ      0
#define CMP_NE      1
#define CMP_GE      2
#define CMP_GT      3
#define CMP_LE      4
#define CMP_LT      5
#define CMP_OR      6
#define CMP_AND     7
#define SHIFT_LEFT  10
#define SHIFT_RIGHT 11

      p = s+i+1;
      if ((s[i] == '=') && (s[i+1] == '=')) {
        p++;
        cmp = CMP_EQ;
      } else if ((s[i] == '!') && (s[i+1] == '=')) {
        p++;
        cmp = CMP_NE;
      } else if ((s[i] == '>') && (s[i+1] == '=')) {
        p++;
        cmp = CMP_GE;
      } else if ((s[i] == '<') && (s[i+1] == '=')) {
        p++;
        cmp = CMP_LE;
      } else if ((s[i] == '&') && (s[i+1] == '&')) {
        p++;
        cmp = CMP_AND;
      } else if ((s[i] == '|') && (s[i+1] == '|')) {
        p++;
        cmp = CMP_OR;
      } else if ((s[i] == '<') && (s[i+1] == '<')) {
        p++;
        cmp = SHIFT_LEFT;
      } else if ((s[i] == '>') && (s[i+1] == '>')) {
        p++;
        cmp = SHIFT_RIGHT;
      } else if (s[i] == '>')
        cmp = CMP_GT;
      else if (s[i] == '<')
        cmp = CMP_LT;

      s[i] = 0;
      value1 = evaluate(eval, s, error);          // eval left side
      if (*error)  return 0;
      value2 = evaluate(eval, p, error);          // eval right side
      if (*error)  return 0;
      switch (cmp) {
      case CMP_EQ:
        if (value1 == value2)  return 1;
        return 0;
        break;
      case CMP_NE:
        if (value1 != value2)  return 1;
        return 0;
        break;
      case CMP_GE:
        if (value1 >= value2)  return 1;
        return 0;
        break;
      case CMP_LE:
        if (value1 <= value2)  return 1;
        return 0;
        break;
      case CMP_GT:
        if (value1 > value2)  return 1;
        return 0;
        break;
      case CMP_LT:
        if (value1 < value2)  return 1;
        return 0;
        break;
      case CMP_OR:
        return ((int)value1) || ((int)value2);
        break;
      case CMP_AND:
        return ((int)value1) && ((int)value2);
        break;
      case SHIFT_LEFT:
        return ((int)value1)<<((int)value2);
        break;
      case SHIFT_RIGHT:
        return ((int)value1)>>((int)value2);
        break;
      }
      return 0;                                   // return result
    }
  }

  // scan for + outside parentheses and convert to (eval) + (eval)
  parentheses = 0;
  for (i = 0; i < length-1; i++) {
    if (s[i] == '(') parentheses++;
    if (s[i] == ')') parentheses--;
    if ((parentheses == 0) && (s[i] == '+')) {
      s[i] = 0;
      value1 = evaluate(eval, s, error);          // eval left side
      if (*error)  return 0;
      value2 = evaluate(eval, s+i+1, error);      // eval right side
      if (*error)  return 0;
      return value1 + value2;                     // return result
    }
  }

  // scan for - outside parentheses and convert to (eval) - (eval)
  parentheses = 0;
  for (i = 0; i < length-1; i++) {
    if (s[i] == '(') parentheses++;
    if (s[i] == ')') parentheses--;
    if ((parentheses == 0) && (s[i] == '-')) {
      s[i] = 0;
      value1 = evaluate(eval, s, error);          // eval left side
      if (*error)  return 0;
      value2 = evaluate(eval, s+i+1, error);      // eval right side
      if (*error)  return 0;
      return value1 - value2;                     // return result
    }
  }

  // scan for * outside parentheses and convert to (eval) * (eval)
  parentheses = 0;
  for (i = 0; i < length-1; i++) {
    if (s[i] == '(') parentheses++;
    if (s[i] == ')') parentheses--;
    if ((parentheses == 0) && (s[i] == '*')) {
      s[i] = 0;                                   // split in left and right side
      value1 = evaluate(eval, s, error);          // eval left side
      if (*error)  return 0;                      // if error, return immediately
      value2 = evaluate(eval, s+i+1, error);      // eval right size
      if (*error)  return 0;                      // if error, return immediately
      return value1 * value2;                     // return result
    }
  }

  // scan for / outside parentheses and convert to (eval) / (eval)
  parentheses = 0;
  for (i = length-1; i >= 0; i--) {               // reversed search direction
    if (s[i] == '(') parentheses++;
    if (s[i] == ')') parentheses--;
    if ((parentheses == 0) && (s[i] == '/')) {
      s[i] = 0;
      value1 = evaluate(eval, s, error);          // eval left side
      if (*error)  return 0;                      // if error, return immediately
      value2 = evaluate(eval, s+i+1, error);      // eval right side
      if ((*error == 0) && (value2 == 0.0)) {
        char errmsg[256];
        sprintf(errmsg, "Division by zero : %s/%s\n", s, s+i+1);
        eval->callback(errmsg);
        *error = EVALUATE_DIVISION_BY_ZERO;
      }
      if (*error)  return 0;
      return value1 / value2;                     // return result
    }
  }

  varno = find_variable(eval, s);
  if (varno >= 0)   return (double)eval->vars[varno].value;

  // check if value is 'PI'
  if ((length == 2) && (s[0] == 'P') && (s[1] == 'I'))  return CONSTANT_PI;

  // check if value is 'e'
  if ((length == 1) && (s[0] == 'e'))  return CONSTANT_e;

  // check for functions XXX()
  // SIN, COS, TAN, EXP, LOG, ABS, SQR
  if (length >= 5) {
    if ((s[3] == '(')  && (s[length-1] == ')') ) {
      func = (s[0]<<16) | (s[1]<<8) | (s[2]);

      if (func == FUNC_SIN) {
        s[length-1] = 0;
        value1 = evaluate(eval, s+4, error);
        if (*error)  return 0;
        return sin(value1);
      }

      if (func == FUNC_COS) {
        s[length-1] = 0;
        value1 = evaluate(eval, s+4, error);
        if (*error)  return 0;
        return cos(value1);
      }

      if (func == FUNC_TAN) {
        s[length-1] = 0;
        value1 = evaluate(eval, s+4, error);
        if (*error)  return 0;
        return tan(value1);
      }

      if (func == FUNC_LOG) {
        s[length-1] = 0;
        value1 = evaluate(eval, s+4, error);
        if (*error)  return 0;
        if (value1 <= 0) {
          char errmsg[256];
          sprintf(errmsg, "log(%s) not allowed\n", s+4);
          eval->callback(errmsg);
          *error = EVALUATE_LOG_ARGUMENT;
          return 0;
        }
        else
          return log(value1);
      }

      if (func == FUNC_EXP) {
        s[length-1] = 0;
        value1 = evaluate(eval, s+4, error);
        if (*error)  return 0;
        return exp(value1);
      }

      if (func == FUNC_ABS) {
        s[length-1] = 0;
        value1 = evaluate(eval, s+4, error);
        if (*error)  return 0;
        return fabs(value1);
      }

      if (func == FUNC_RND) {
        s[length-1] = 0;
        value1 = evaluate(eval, s+4, error);
        if (*error)  return 0;
        if (value1 <= 0) {
          char errmsg[256];
          sprintf(errmsg, "rnd(%s) not allowed\n", s+4);
          eval->callback(errmsg);
          *error = EVALUATE_RND_ARGUMENT;
          return 0;
        }
        return (double)(rand() % (int)value1);
      }

      if (func == FUNC_SQR) {
        s[length-1] = 0;
        value1 = evaluate(eval, s+4, error);
        if (*error)  return 0;
        if (value1 < 0) {
          char errmsg[256];
          sprintf(errmsg, "sqr(%s) not allowed\n", s+4);
          eval->callback(errmsg);
          *error = EVALUATE_SQR_ARGUMENT;
          return 0;
        }
        else
          return sqrt(value1);
      }

    }
  }

  // couldn't find neither + - * / so the rest must be a value
  if (sscanf(s, "%e", &value4) != 1) {
    char errmsg[256];
    sprintf(errmsg, "Failed to evaluate '%s'\n", s);
    eval->callback(errmsg);
    *error = EVALUATE_FAILED;
    return 0;
  }

  return value4;
}


void dummy_callback(char *err) {
}
