#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <math.h>
#include <ctype.h>

#include "proto.h"
#include "preprocess.h"



struct PREPROCESS {
  MACRO macros[MAXMACROS];
  S32 macrocount;

  LINE *lines;
  int linecount, allocatedlines;

  char *filenames[MAXFILES];
  char include_at_tail[MAXFILES];
  int filenamecount;

  U32 allocated, codesize, ifdefdepth;
  char *buffer;

  void (* callback)(char *err);

  char *errmsg;
  char libdir[256];
  char basedir[256];
};


static S32 insert_file(PREPROCESS *p, char *dir, char *filename, int fileno);
static S32 macro_substitute(PREPROCESS *p, char *line, char *out);
static S32 insert_line(PREPROCESS *p, char *line, int lineno, int fileno);
static S32 define_macro(PREPROCESS *p, char *name, char *value);

static S32 is_macro_start_char(char s);
static S32 is_macro_char(char s);
static void strip_leading_trailing_spaces(char *in, char **out);
static void build_filename(char *out, char *dir, char *filename);
static void convert_from_C_to_native(char *filename);


EXPORT_C PREPROCESS *preprocessor(char *dir, char *libdir, char *filename, char *errmsg) {

  S32 err, i;
  PREPROCESS *pre;

  pre = malloc(sizeof(PREPROCESS));
  if (!pre)   return NULL;
  memset(pre, 0, sizeof(PREPROCESS));

  pre->errmsg = errmsg;
  *errmsg = '\0';

  pre->buffer = malloc(32*1024);
  if (!pre->buffer) {
    free(pre);
    strcpy(errmsg, "No room");
    return NULL;
  }
  pre->allocated = 32*1024;
  pre->codesize = 0;
  pre->ifdefdepth = 0;
  pre->filenamecount = 0;
  pre->linecount = pre->allocatedlines = 0;
  pre->macrocount = 0;
  strcpy(pre->libdir, libdir);
  strcpy(pre->basedir, dir);

  err = insert_file(pre, dir, filename, 0);
  if (err)   return NULL;

  for (i = 0; i < MAXFILES; i++)
    if (pre->include_at_tail[i]) {
      err = insert_file(pre, "", pre->filenames[i], i);
      if (err)   return NULL;
    }

  for (i = 0; i < pre->linecount; i++)   pre->lines[i].line += (int)pre->buffer;

  return pre;
}


EXPORT_C void preprocessor_kill(PREPROCESS *pre) {

  int i;

  for (i = 0; i < pre->filenamecount; i++) {
    free(pre->filenames[i]);
    pre->filenames[i] = NULL;
  }

  if (pre->buffer)   free(pre->buffer);
  pre->buffer = NULL;

  if (pre->lines)    free(pre->lines);
  pre->lines = NULL;

  free(pre);
}


EXPORT_C void preprocessor_info(PREPROCESS *pre, LINE **l, int *lc, char ***f, int *fc) {

  *l  = pre->lines;
  *lc = pre->linecount;
  *f  = pre->filenames;
  *fc = pre->filenamecount;
}


EXPORT_C S32 find_macro(PREPROCESS *pre, char *name) {

  S32 i;
  char *p;

  while (isspace(name[0]))     name++;
  p = name;
  while (is_macro_char(*p))  p++;
  *p = '\0';

  for (i = 0; i < pre->macrocount; i++)
     if (strcmp(pre->macros[i].name, name) == 0)  return i;

  return -1;
}

// --------------------------------------------------------------------


S32 insert_file(PREPROCESS *pre, char *dir, char *filename, int fileno) {

  FILE *fh;
  char linep[MAXLINELENGTH], fullfilename[256];
  int lineno;

  if (fileno > MAXFILES)   return 1;
  if (fileno < pre->filenamecount)
    // file is already in the filelist, so it has also been converted to native format
    strcpy(fullfilename, filename);
  else {
    // add filename to list of filenames
    pre->filenames[fileno] = malloc(strlen(filename)+1);
    if (!pre->filenames[fileno])   return 1;
    strcpy(pre->filenames[fileno], filename);
    pre->filenamecount++;
    // convert to native format
    build_filename(fullfilename, dir, filename);
  }

  // open the file
  fh = fopen(fullfilename, "r");
  if (!fh) {
    sprintf(pre->errmsg, "Failed to open file '%s'\n", filename);
    return 1;
  }

  lineno = 0;
  do {
    if (fgets(linep, MAXLINELENGTH, fh)) {
      S32 i;
      char *line;

      lineno++;
      // convert TAB, linefeed etc to normal spaces
      line = linep;
      line[MAXLINELENGTH-1] = '\0';
      while ((*line) && (*line < ' '))  line++;
      for(i = 0; i < strlen(line); i++)
        if (line[i] < ' ')  line[i] = ' ';

      if (line[0] == '#') {
        if (strncmp(line, "#define ", 8) == 0) {
          if (!pre->ifdefdepth) {
            char temp[MAXLINELENGTH], *value, *name;

            name = line+8;
            while (*name == ' ')        name++;     // skip to macro name
            value = name;
            while (is_macro_char(*value))  value++; // skip to end of macro name
            *value++ = '\0';                        // terminate name
            while (*value == ' ')        value++;   // skip to macro name
            if (macro_substitute(pre, value, temp)) {
              fclose(fh);
              return 1;
            }
            if (define_macro(pre, name, temp)) {
              fclose(fh);
              return 1;
            }
          }

        } else if (strncmp(line, "#undef ", 7) == 0) {
          if (!pre->ifdefdepth) {
            S32 mi;
            mi = find_macro(pre, line+7);
            if (mi >= 0) {
              if (mi != pre->macrocount-1) {
                pre->macros[mi].name  = pre->macros[pre->macrocount-1].name;
                pre->macros[mi].value = pre->macros[pre->macrocount-1].value;
              }
              pre->macrocount--;
            }
          }

        } else if (strncmp(line, "#ifdef ", 7) == 0) {
          if (pre->ifdefdepth)
            pre->ifdefdepth++;
          else {
            if (find_macro(pre, line+7) < 0)  pre->ifdefdepth++;
          }

        } else if (strncmp(line, "#else", 5) == 0) {
          if (pre->ifdefdepth == 1)
            pre->ifdefdepth = 0;
          else if (pre->ifdefdepth == 0)
            pre->ifdefdepth = 1;

        } else if (strncmp(line, "#ifndef ", 8) == 0) {
          if (pre->ifdefdepth)
            pre->ifdefdepth++;
          else {
            if (find_macro(pre, line+8) >= 0)  pre->ifdefdepth++;
          }

        } else if (strncmp(line, "#endif", 6) == 0) {
          if (pre->ifdefdepth)  pre->ifdefdepth--;

        } else if (strncmp(line, "#include ", 9) == 0) {
          if (!pre->ifdefdepth) {
            char *filename, *p, *searchpath;

            strip_leading_trailing_spaces(line+9, &filename);
            p = filename + strlen(filename);
            if (p != filename) {
              p--;
              if ((*p == '\"') && (*filename == '\"')) {
                filename++;
                *p = '\0';
                searchpath = dir;
              } else if ((*p == '>') && (*filename == '<')) {
                filename++;
                *p = '\0';
                searchpath = pre->libdir;
              } else {
                searchpath = "";
              }
            }
            if (insert_file(pre, searchpath, filename, fileno+1)) {
              fclose(fh);
              return 1;
            }
          }

        } else if (strncmp(line, "#tail ", 6) == 0) {
          if (!pre->ifdefdepth) {
            char *filename, *p, *searchpath;

            strip_leading_trailing_spaces(line+6, &filename);
            p = filename + strlen(filename);
            if (p != filename) {
              p--;
              if ((*p == '\"') && (*filename == '\"')) {
                filename++;
                *p = '\0';
                searchpath = dir;
              } else if ((*p == '>') && (*filename == '<')) {
                filename++;
                *p = '\0';
                searchpath = pre->libdir;
              } else {
                searchpath = "";
              }
            }

            pre->filenames[fileno+1] = malloc(256);
            if (!pre->filenames[fileno+1]) {
              fclose(fh);
              return 1;
            }
            build_filename(pre->filenames[fileno+1], searchpath, filename);
            pre->include_at_tail[fileno+1] = 1;
            pre->filenamecount++;
          }

        } else if (!pre->ifdefdepth) {
          if (insert_line(pre, line, lineno, fileno)) {
            fclose(fh);
            return 1;
          }
        }

      } else if (!pre->ifdefdepth) {
        char output[MAXLINELENGTH];

        if (macro_substitute(pre, line, output)) {
          fclose(fh);
          return 1;
        }
        if (insert_line(pre, output, lineno, fileno)) {
          fclose(fh);
          return 1;
        }
      }

    } else {
      fclose(fh);
      return 0;
    }
  } while (fh);

  return 1;
}


S32 insert_line(PREPROCESS *pre, char *line, int lineno, int fileno) {

  S32 len;

  len = strlen(line)+1;
  if (pre->codesize + len >= pre->allocated)  {
    char *newbuffer;

    newbuffer = realloc(pre->buffer, pre->allocated + 32*1024);
    if (!newbuffer)  return 1;
    pre->allocated += 32*1024;
    pre->buffer = newbuffer;
  }
  memcpy(pre->buffer+pre->codesize, line, len);

  if (pre->linecount == pre->allocatedlines) {
    if (pre->allocatedlines)
      pre->lines = realloc(pre->lines, (pre->allocatedlines+512)*sizeof(LINE));
    else
      pre->lines = malloc(512*sizeof(LINE));
    if (!pre->lines)   return 1;
    pre->allocatedlines += 512;
  }
  pre->lines[pre->linecount].line = (char *)pre->codesize;
  pre->lines[pre->linecount].lineno = lineno;
  pre->lines[pre->linecount].fileno = fileno;

  pre->linecount++;
  pre->codesize += len;

  return 0;
}


S32 define_macro(PREPROCESS *pre, char *name, char *value) {

  S32 mi;
  char *namep, *valuep;

  if (pre->macrocount == MAXMACROS) {
    strcpy(pre->errmsg, "Too many macros");
    return 1;
  }

  namep = malloc(strlen(name)+1);
  if (!namep) {
    strcpy(pre->errmsg, "No room for macro");
    return 1;
  }
  valuep = malloc(strlen(value)+1);
  if (!valuep) {
    free(namep);
    strcpy(pre->errmsg, "Too many macros");
    return 1;
  }

  strcpy(namep, name);
  strcpy(valuep, value);
  if (*valuep) {
    char *end;

    end = valuep + strlen(valuep) - 1;
    while (*end == ' ')   *end-- = '\0';
  }

  mi = find_macro(pre, namep);
  if (mi != -1) {
    // macro already defined
    if (strcmp(valuep, pre->macros[mi].value) == 0)  return 0;
    sprintf(pre->errmsg, "Differing redefinition of macro '%s'", namep);
    return 1;
  }

  pre->macros[pre->macrocount].name = namep;
  pre->macros[pre->macrocount].value = valuep;

  pre->macrocount++;
  return 0;
}


S32 macro_substitute(PREPROCESS *pre,char *line, char *out) {

  S32 rd, wr, len, checkmacro;

  len = strlen(line);
  checkmacro = 1;
  wr = 0;
  for (rd = 0; rd < len; rd++) {
    // don't do macro-substitution inside ""
    if (line[rd] == '"') {
      S32 quotes;
      out[wr++] = line[rd++];
      quotes = 1;
      while (quotes && (line[rd])) {
        if (line[rd] == '"')  quotes = 0;
        out[wr++] = line[rd++];
      }
    }
    if ((checkmacro) && (is_macro_start_char(line[rd]))) {
      S32 m;
      for (m = 0; m < pre->macrocount; m++) {
        char *macro, *value;
        S32 len;
        macro = pre->macros[m].name;
        value = pre->macros[m].value;
        len = strlen(macro);
        if ((strncmp(line+rd, macro, len) == 0) && (!is_macro_char(line[rd+len]))) {
          if (wr + strlen(value) > MAXLINELENGTH-1) {
            sprintf(pre->errmsg, "Macro-substituted line is too long");
            return 1;
          }
          strcpy(out+wr, value);
          wr += strlen(value);
          rd += strlen(macro);
          m = pre->macrocount+1;         // abort
          rd--;
        }
      }
      if (m == pre->macrocount)  out[wr++] = line[rd];
      checkmacro = 0;

    } else {
      if (!is_macro_char(line[rd]))
        checkmacro = 1;
      else
        checkmacro = 0;
      out[wr++] = line[rd];
    }
  }
  out[wr] = '\0';

  return 0;
}


void strip_leading_trailing_spaces(char *in, char **out) {

  while (*in == ' ')  in++;
  *out = in;
  if (!*in)                   return;

  in += strlen(in)-1;
  while (*in == ' ')  *in = '\0', in--;
}


S32 is_macro_start_char(char s) {

  if (isalpha(s) || (s == '_'))     return 1;
  return 0;
}


S32 is_macro_char(char s) {

  if (isalpha(s) || isdigit(s) || (s == '_'))     return 1;
  return 0;
}


void build_filename(char *out, char *dir, char *name) {

  char filename[256];

  strcpy(filename, name);
  convert_from_C_to_native(filename);

  if (*dir)
#ifdef __riscos__
    sprintf(out, "%s.%s", dir, filename);
#else
#ifdef __epoc32__
    sprintf(out, "%s\\%s", dir, filename);
#else
    sprintf(out, "%s/%s", dir, filename);
#endif
#endif
  else
    strcpy(out, filename);
}


void convert_from_C_to_native(char *filename) {

  int i;

  i = 0;
#ifdef __riscos__
  while (filename[i]) {
    if (filename[i] == '.')
      filename[i] = '/';
    else if (filename[i] == '/')
      filename[i] = '.';
    i++;
  }
#endif
#ifdef __epoc32__
  while (filename[i]) {
    if (filename[i] == '/')
      filename[i] = '\\';
    else if (filename[i] == '\\')
      filename[i] = '/';
    i++;
  }
#endif

}
